import time, sys, socket


# ------------------------------------------------------
class ObjectTree:

	map_Objects = {}

	#														
	def __init__ (self, sID, sMainID, sActID, sObject):
		self.s_ID = sID
		self.s_MainID = sMainID
		self.s_ActID = sActID
		self.s_Object = sObject
		self.map_Children = {}


	#														
	def add (self, sID, sMainID, sActID, sObject):
		oTree = ObjectTree (sID, sMainID, sActID, sObject)
		ObjectTree.map_Objects [sID] = oTree


	#														
	def build_tree (self):
		for oTree in ObjectTree.map_Objects.values ():
			if oTree.s_MainID == oTree.s_ID:
				self.map_Children [oTree.s_Object + '\n' + oTree.s_ID] = oTree
				continue

			oParent = ObjectTree.map_Objects.get (oTree.s_ActID, None)
			if None != oParent:
				oParent.map_Children [oTree.s_Object + '\n' + oTree.s_ID] = oTree
				continue

			oParent = ObjectTree.map_Objects.get (oTree.s_MainID, None)
			if None != oParent:
				oParent.map_Children [oTree.s_Object + '\n' + oTree.s_ID] = oTree
				continue

			self.map_Children [oTree.s_Object + '\n' + oTree.s_ID] = oTree
		ObjectTree.map_Objects.clear ()


	#														
	def to_text (self, _iLevel):
		sText = ''

		lstIDs = self.map_Children.keys ()
		lstIDs.sort ()
		for sID in lstIDs:
			oTree = self.map_Children [sID]
			sText = sText + ''.join ([' '] * _iLevel) + oTree.s_Object + '\n'
			sChildren = oTree.to_text (_iLevel + 1)
			if '' != sChildren:
				sText = sText + sChildren + '\n'

		return sText.strip ()



#														
def convert_to_normalized_response (_sResponse):
	lstRawObjects = _sResponse.split ('\n')
	# oTreeRoot = ObjectTree ('', '', '', '')

	lstObjects = []
	for sObject in lstRawObjects:
		if '\x01' in sObject:
			lstObjects.append (sObject.strip ())
			# (sID, sMainID, sActID, sObject) = sObject.split ('\x01', 3)
			# oTreeRoot.add (sID.strip (), sMainID.strip (), sActID.strip (), sObject.strip ())
	# oTreeRoot.build_tree ()
	# sGenericResponse = oTreeRoot.to_text (0)
	lstObjects.sort ()
	sGenericResponse = '\n'.join (lstObjects)
	return sGenericResponse.strip ()



# ------------------------------------------------------
class CacheNode:

	#														
	def __init__ (self):
		self.s_Response = None
		self.s_NormalizedResponse = None
		self.map_ActionToResponse = {}


	#														
	def clean (self, _sMsg):
		for sMsg, oCacheNode in self.map_ActionToResponse.items ():
			if ('restart complete' in oCacheNode.s_Response) or \
				('ERROR' in oCacheNode.s_Response):
				del self.map_ActionToResponse [sMsg]
				Cache.i_NodesCleaned = Cache.i_NodesCleaned + 1
				continue
			oCacheNode.clean (_sMsg + sMsg)


	#														
	def trim (self):
		for sMsg, oCacheNode in self.map_ActionToResponse.items ():
			if len (oCacheNode.map_ActionToResponse) == 0:
				del self.map_ActionToResponse [sMsg]
				continue
			oCacheNode.trim ()


	#														
	def save (self, _file, _mapNodeToLinks, _mapNodeToID):
		if self in _mapNodeToID:
			return
		id = len (_mapNodeToID)
		_mapNodeToID [self] = str (id)

		_file.write ('<><><><><><><><><><><>\n')
		_file.write (str (id) + '\n')
		if None != self.s_Response:
			_file.write (self.s_Response + '\n')
		else:
			_file.write ('\n')

		_mapNodeToLinks [self] = self.map_ActionToResponse
		for (action, state) in self.map_ActionToResponse.items ():
			state.save (_file, _mapNodeToLinks, _mapNodeToID)





# ------------------------------------------------------
class Cache:

	i_NodesCleaned = 0

	#														
	def __init__ (self):
		self.p_Root = CacheNode ()
		self.map_NomalizedResponseToNode = {}


	#														
	def get_node (self, _sMessage, _sResponse):
		if None == _sResponse:
			print '[ERROR]  cache.get_node called with sResponse = None !'
			file_InteractionLog.write ('[ERROR]  cache.get_node called with sResponse = None !\n')
			assert False

		if _sMessage.startswith ('*'):
			sNormalizedResponse = convert_to_normalized_response (_sResponse)
			oNode = self.map_NomalizedResponseToNode.get (sNormalizedResponse, None)
			if None == oNode:
				oNode = CacheNode ()
				oNode.s_Response = _sResponse
				oNode.s_NormalizedResponse = sNormalizedResponse
				self.map_NomalizedResponseToNode [sNormalizedResponse] = oNode
			return oNode
		else:
			oNode = CacheNode ()
			oNode.s_Response = _sResponse
			oNode.s_NormalizedResponse = _sResponse
			return oNode



	#														
	def load (self, _file):
		# read cache nodes ...
		sID = None
		sResponse = ''

		map_IdToNode = {}

		sLine = None
		while True:
			sLine = _file.readline ()
			if '' == sLine:
				break
			sLine = sLine.strip ()
			if '' == sLine:
				continue
			if '<<>><<>><<>>' in sLine:
				if None != sID:
					node = CacheNode ()
					node.s_Response = sResponse
					node.s_NormalizedResponse = convert_to_normalized_response (node.s_Response)
					map_IdToNode [sID] = node
				break

			if '<><><><><>' in sLine:
				if None != sID:
					node = CacheNode ()
					node.s_Response = sResponse
					node.s_NormalizedResponse = convert_to_normalized_response (node.s_Response)
					map_IdToNode [sID] = node

				sID = None
				sResponse = ''
				continue

			if None == sID:
				sID = sLine
				continue
			if '' != sResponse:
				sResponse = sResponse + '\n'
			sResponse = sResponse + sLine
			

		self.p_Root = map_IdToNode ['0']


		# read node links ...
		sFrom = ''
		sLine = None
		while True:
			sLine = _file.readline ()
			if '' == sLine:
				break

			sLine = sLine.strip ()
			if '' == sLine:
				continue
			if sLine.startswith ('from : '):
				sFrom = sLine [7:].strip ()
				continue
			if sLine.startswith ('to : '):
				(sTo, sMessage) = sLine [5:].strip ().split ('\x09')
				sTo = sTo.strip ()
				sMessage = sMessage.strip ()

				# print sFrom + ' -> ' + sTo
				nodeFrom = map_IdToNode.get (sFrom, None)
				if None == nodeFrom:
					print 'missing node : ' + sFrom
					print '        to   : ' + sTo
					print '     message : ' + sMessage
				assert (None != nodeFrom)
				nodeTo = map_IdToNode.get (sTo, None)
				if None == nodeTo:
					print 'missing node : ' + sFrom
					print '        to   : ' + sTo
					print '     message : ' + sMessage
				assert (None != nodeTo)

				nodeFrom.map_ActionToResponse [sMessage.strip ()] = nodeTo

		print '      cache : loaded ' + str (len (map_IdToNode)) + ' states.'



	#														
	def save (self, _file):
		map_NodeToLinks = {}
		map_NodeToID = {}

		self.p_Root.save (_file, map_NodeToLinks, map_NodeToID)
		
		_file.write ('<<>><<>><<>><<>><<>><<>>\n')
		for node_from, mapLinks in map_NodeToLinks.items ():
			_file.write ('from : ' + map_NodeToID [node_from] + '\n')
			for msg, node_to in mapLinks.items ():
				if None == map_NodeToID.get (node_to, None):
					print map_NodeToID [node_from]
				_file.write ('to : ' + map_NodeToID [node_to] + '\x09' + msg.strip () + '\n')



	# ======================================================
	# Peripheral functions									
	# ======================================================


	#														
	def check (self, _sMessage):
		oCacheNode = self.p_CurrentNode.map_ActionToResponse.get (_sMessage.strip (), None)
		if None != oCacheNode:
			self.lst_Messages.append (_sMessage)
			self.p_CurrentNode = oCacheNode
			return oCacheNode.s_Response

		else:
			return '[dead end]\n'


	#														
	#def clean (self):
	#	for oStartingNode in self.p_Root.map_ActionToResponse.values ():
	#		for sMsg, oCacheNode in oStartingNode.map_ActionToResponse.items ():
	#			if ('restart complete' in oCacheNode.s_Response) or \
	#				('ERROR' in oCacheNode.s_Response):
	#				del self.p_Root.map_ActionToResponse [sMsg]
	#				Cache.i_NodesCleaned = Cache.i_NodesCleaned + 1
	#				continue
	#			oCacheNode.clean (sMsg)
	#	print '   cleaned ' + str (Cache.i_NodesCleaned) + ' cache nodes.'


	#														
	#def trim (self):
	#	for oCacheNode in self.p_Root.map_ActionToResponse.values ():
	#		oCacheNode.trim ()
	#	print '   cache leaves trimmed.'


