
	def word_maybe_object (self, _sWord, obj):
		dDistance = normalized_alnum_edit_distance (obj.s_CleanName, _sWord)

		# if the object has low edit distance, we don't care about substring
		if dDistance < self.d_EditDistCutoff:
			return True
		# if the object has high edit distance, we don't care about substring
		if dDistance < self.d_SubStrEditDistCutoff:
			# check substring
			if obj.s_CleanName.startswith (_sWord + ' '):
				return True
			elif obj.s_CleanName.endswith (' ' + _sWord):
				return True
			elif ' ' + _sWord + ' ' in obj.s_CleanName:
				return True
			elif _sWord == obj.s_CleanName:
				return True
		return False


	# check if the environment has an object with the current word	
	def current_env_has_word (self, _sWord):
		bWordInEnv = self.map_WordsInEnvironment.get (_sWord, None)
		if None != bWordInEnv:
			return bWordInEnv

		for obj in self.lst_CurrentState:
			if True == self.word_maybe_object (_sWord, obj):
				self.map_WordsInEnvironment [_sWord] = True
				return True
		self.map_WordsInEnvironment [_sWord] = False
		return False



	#																
	def current_env_has_word_approx (self, _sWord):
		for obj in self.lst_CurrentState:
			dDistance = normalized_alnum_edit_distance (obj.s_Name, _sWord)
			if dDistance < self.d_EditDistCutoff:
				return True
		return False



	#																
	def current_env_has_word_substr (self, _sWord):
		for obj in self.lst_CurrentState:
			if obj.s_CleanName.startswith (_sWord + ' '):
				return True
			elif obj.s_CleanName.endswith (' ' + _sWord):
				return True
			elif ' ' + _sWord + ' ' in obj.s_CleanName:
				return True
			elif _sWord == obj.s_CleanName:
				return True
		return False



	#																
	def distance_to_index (self, iDistance):
		if iDistance < 0:
			iDistance = iDistance + 10
			if iDistance < 0:
				iDistance = 0
		else:
			iDistance = iDistance + 9
			if iDistance > 19:
				iDistance = 19
		return iDistance



	# compute the lowest edit distance between sWord and the 	
	# names of the objects in the environment.					
	def lowest_edit_distance_in_env_to_word (self, _sWord):
		dLowestEditDistance = None
		for obj in self.lst_CurrentState:
			dDistance = normalized_alnum_edit_distance (obj.s_Name, _sWord)
			if (None == dLowestEditDistance) or (dDistance < dLowestEditDistance):
				dLowestEditDistance = dDistance
		return dDistance



	# features used for selecting the word indicating the UI object		
	def feature_wc_wu (self, _iLocationWC, _iLocationWU, _dWeights):
		fv = [0.0] * self.i_FeatureCount_WC_WU
		dScore = 0.0

		# ------------------------------------------------------
		# WC features ...										
		# ------------------------------------------------------

		# current word's type									
		iIndexWC = self.lst_CurrentInstructionWordIndices [_iLocationWC]
		fv [iIndexWC] = 1.0
		dScore += _dWeights [iIndexWC]
		iOffset = self.i_VocabularySize

		# distance of current word from last used word			
		iDistanceToU = _iLocationWC + 1
		if len (self.lst_PreviouslyUsedObjectWords) > 0:
			iDistanceToU = _iLocationWC - self.lst_PreviouslyUsedObjectWords [-1]
		iDistanceToC = _iLocationWC + 1
		if len (self.lst_PreviouslyUsedCommandWords) > 0:
			iDistanceToC = _iLocationWC - self.lst_PreviouslyUsedCommandWords [-1]
		iDistanceToP = 0
		if len (self.lst_PreviouslyUsedParamWords) > 0:
			iDistanceToP = _iLocationWC - self.lst_PreviouslyUsedParamWords [-1]

		iDistance = None
		if (iDistanceToU > 0) and (iDistanceToC > 0):
			iDistance = min (iDistanceToU, iDistanceToC)
			if iDistanceToP > 0:
				iDistance = min (iDistance, iDistanceToP)
		if (iDistanceToU < 0) and (iDistanceToC < 0):
			iDistance = max (iDistanceToU, iDistanceToC)
			if iDistanceToP < 0:
				iDistance = max (iDistance, iDistanceToP)

		if None != iDistance:
			iDistanceIndex = self.distance_to_index (iDistance)
			fv [iOffset + iDistanceIndex] = 1.0
			dScore += _dWeights [iOffset + iDistanceIndex]
		iOffset += 20

		# ------------------------------------------------------
		# WU features ...										
		# ------------------------------------------------------
		iIndexWU = self.lst_CurrentInstructionWordIndices [_iLocationWU]
		sWordWU = self.lst_CurrentInstructionWords [_iLocationWU]
		bObjectWordNone = ('[[none]]' == sWordWU)

		if True == self.b_UseBigramFeatures:
			# [previous WU] [current WU] bigram						
			if False == bObjectWordNone:
				iPreviousWU = 0
				if len (self.lst_PreviouslyUsedObjectWords) > 0:
					iPreviousWU = self.lst_CurrentInstructionWordIndices [self.lst_PreviouslyUsedObjectWords [-1]]
				fv [iOffset + iPreviousWU * self.i_VocabularySize + iIndexWU] = 1.0
				dScore += _dWeights [iOffset + iPreviousWU * self.i_VocabularySize + iIndexWU]
			iOffset += self.i_VocabularySize * self.i_VocabularySize

		# presence of objects in s^*_t containing current word	
		# lowest edit distance in s^*_t to current word			
		if False == bObjectWordNone:
			if True == self.current_env_has_word (sWordWU):
				fv [iOffset] = 1.0
				dScore += _dWeights [iOffset]
				fv [iOffset + 1] = self.lowest_edit_distance_in_env_to_word (sWordWU)
				dScore += fv [iOffset + 1] * _dWeights [iOffset + 1]
		iOffset += 2

		#-------------------------------------------------------
		# WC - WU joint features								
		#-------------------------------------------------------

		# distance of WU from WC								
		if False == bObjectWordNone:
			iDistance = _iLocationWU - _iLocationWC
			iDistanceIndex = self.distance_to_index (iDistance)
			fv [iOffset + iDistanceIndex] = 1.0
			dScore += _dWeights [iOffset + iDistanceIndex]
		iOffset += 20

		if True == self.b_UseBigramFeatures:
			# [WC] [WU] bigram										
			iWC = self.lst_CurrentInstructionWordIndices [_iLocationWC]
			fv [iOffset + iWC * self.i_VocabularySize + iIndexWU] = 1.0
			dScore += _dWeights [iOffset + iWC * self.i_VocabularySize + iIndexWU]
			iOffset += self.i_VocabularySize * self.i_VocabularySize


		#														
		assert (iOffset == self.i_FeatureCount_WC_WU)
		self.d_Features_WC_WU [_iLocationWC][_iLocationWU] = fv
		return dScore



	# features of object												
	# this is called only for objects containing WU in their name		
	def feature_u (self, _iLocationWU, _oU, _dWeights):
		# features of object									
		fv = [0.0] * self.i_FeatureCount_U

		dScore = 0
		for f, m in self.map_FeatureIndexToMessageIndex.items ():
			if NAME_EDIT_DIST == f:
				continue
			fv [f] = _oU.d_ReplyFeatures [m]
			dScore += _dWeights [f] * fv [f]

		# edit distance of WU to object (needs to be recomputed)
		dEditDistance = normalized_alnum_edit_distance (\
							_oU.s_Name, self.lst_CurrentInstructionWords [_iLocationWU])
		fv [NAME_EDIT_DIST] = dEditDistance
		dScore += _dWeights [NAME_EDIT_DIST] * dEditDistance

		# presence of object's class name in neighbourhood of WU
		iLeft = _iLocationWU - self.i_Neighbourhood_U
		if iLeft < 0:
			iLeft = 0
		iRight = _iLocationWU + self.i_Neighbourhood_U
		if iRight > self.i_CurrentInstructionLength:
			iRight = self.i_CurrentInstructionLength
		if _oU.s_ObjectClass in self.lst_CurrentInstructionWords [iLeft : iRight]:
			fv [-1] = 1.0
			dScore += _dWeights [-1]

		#														
		self.d_Features_U [_iLocationWU][_oU.i_Index] = fv
		return dScore



	# features used for selecting the UI command						
	def feature_c (self, _oU, _iLocationWU, _iLocationWC, _iC, _dWeights):
		fv = [0.0] * self.i_FeatureCount_C
		dScore = 0.0

		# WC and neighbouring words 							
		iwc = self.lst_CurrentInstructionWordIndices [_iLocationWC]
		fv [iwc + _iC * self.i_VocabularySize] = 1.0
		dScore += _dWeights [iwc + _iC * self.i_VocabularySize]
		iOffset = self.i_CommandCount * self.i_VocabularySize

		if _iLocationWC > 0:
			iwc = self.lst_CurrentInstructionWordIndices [_iLocationWC - 1]
			#fv [iOffset + iwc + _iC * self.i_VocabularySize] = 1
			#dScore += _dWeights [iOffset + iwc + _iC * self.i_VocabularySize]

		if _iLocationWC < self.i_CurrentInstructionLength - 1:
			iwc = self.lst_CurrentInstructionWordIndices [_iLocationWC + 1]
			#fv [iOffset + iwc + _iC * self.i_VocabularySize] = 1
			#dScore += _dWeights [iOffset + iwc + _iC * self.i_VocabularySize]
		#iOffset += self.i_CommandCount * self.i_VocabularySize

		# UI object's class										
		if None != _oU.i_ObjectClass:
			fv [iOffset + _oU.i_ObjectClass + _iC * self.i_ObjectClassCount] = 1.0
			dScore += _dWeights [iOffset + _oU.i_ObjectClass + _iC * self.i_ObjectClassCount]
		iOffset += self.i_CommandCount * self.i_ObjectClassCount

		# Object word & command 								
		iwu = self.lst_CurrentInstructionWordIndices [_iLocationWU]
		fv [iOffset + iwu + _iC * self.i_VocabularySize] = 1.0
		dScore += _dWeights [iOffset + iwu + _iC * self.i_VocabularySize]
		iOffset += self.i_VocabularySize * self.i_CommandCount

		#														
		self.d_Features_C [_iLocationWC][_iLocationWU][_oU.i_Index][_iC] = fv
		assert (iOffset == self.i_FeatureCount_C)
		return dScore



	#																	
	def feature_wp (self, _iLocationWU, _iLocationWC, _iWordLocation, _dWeights):
		fv = [0.0] * self.i_FeatureCount_WP
		dScore = 0.0

		# the word itself										
		iIndexWP = self.lst_CurrentInstructionWordIndices [_iWordLocation]
		fv [iIndexWP] = 1.0
		dScore += _dWeights [iIndexWP]
		iOffset = 1

		# words in the neighbourhood of current word			
		iLeft = _iWordLocation - self.i_Neighbourhood_WP
		if iLeft < 0:
			iLeft = 0
		for w in self.lst_CurrentInstructionWordIndices [iLeft : _iWordLocation]:
			fv [iOffset + w] = 1.0
			dScore += _dWeights [iOffset + w]

		iRight = _iWordLocation + self.i_Neighbourhood_WP
		if iRight > self.i_CurrentInstructionLength:
			iRight = self.i_CurrentInstructionLength
		for w in self.lst_CurrentInstructionWordIndices [_iWordLocation + 1 : iRight]:
			fv [iOffset + w] = 1.0
			dScore += _dWeights [iOffset + w]

		iOffset += self.i_VocabularySize

		# distance of current word from WU						
		iDistance = _iWordLocation - _iLocationWU
		iDistanceIndex = self.distance_to_index (iDistance)
		fv [iOffset + iDistanceIndex] = 1.0
		dScore += _dWeights [iOffset + iDistanceIndex]
		iOffset += 20

		# distance of current word from WC						
		iDistance = _iWordLocation - _iLocationWC
		iDistanceIndex = self.distance_to_index (iDistance)
		fv [iOffset + iDistanceIndex] = 1.0
		dScore += _dWeights [iOffset + iDistanceIndex]
		iOffset += 20

		#														
		assert (iOffset == self.i_FeatureCount_WP)
		self.d_Features_WP [_iWordLocation] = fv
		return dScore



	#															
	def compute_score_wc_wu (self, _iLocationWC, _iLocationWU):
		dScore = self.feature_wc_wu (_iLocationWC, _iLocationWU, self.d_Weights_WC_WU)
		dScore = dScore - self.d_AverageWeights_WC_WU
		return dScore * self.d_ExplorationRate


	#															
	def compute_score_u (self, _iLocationWU, _oU):
		dScore = self.feature_u (_iLocationWU, _oU, self.d_Weights_U)
		dScore = dScore - self.d_AverageWeights_U
		return dScore * self.d_ExplorationRate


	#															
	def compute_score_c (self, _oU, _iLocationWU, _iLocationWC, _iCommandIndex):
		dScore = self.feature_c (_oU, _iLocationWU, _iLocationWC, _iCommandIndex, self.d_Weights_C)
		dScore = dScore - self.d_AverageWeights_C
		return dScore * self.d_ExplorationRate


	#															
	def compute_score_wp (self, _iLocationWU, _iLocationWC, _iWordLocation):
		dScore = self.feature_wp (_iLocationWU, _iLocationWC, _iWordLocation, self.d_Weights_WP)
		dScore = dScore - self.d_AverageWeights_WP
		if dScore > 700:
			dScore = 700
		return dScore * self.d_ExplorationRate



